"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Logger = exports.LOG_DELIMITER = void 0;
const events_1 = require("events");
const fast_safe_stringify_1 = require("fast-safe-stringify");
const strings_1 = require("./strings");
const system = require("./system");
const types_1 = require("./types");
/*
Properties of an ideal logger
  accepts variadic arguments
  nicely formats messages
  severity suppression
  redaction
  sub contexts
  allows custom formatting of objects
*/
exports.LOG_DELIMITER = '___$LAMBDA-LOG-TAG$___';
const reservedKeys = ['message', 'severity'];
function Logger({ minimumLogLevel, formatter, useBearerRedactor = true, useGlobalErrorHandler = true, forceGlobalErrorHandler = false, redactors = [], testMode = undefined, } = {}) {
    // This wacky omit it just to make typescript happy regarding the props
    // that are initialized later in this function
    const initContext = {
        minimumLogLevel,
        formatter: (formatter !== null && formatter !== void 0 ? formatter : isInTestMode(testMode)) ? TestFormatter : JsonFormatter,
        events: new events_1.EventEmitter(),
        contextPath: [],
        testMode,
        keys: new Map(),
    };
    if (useBearerRedactor) {
        redactors.push(bearerRedactor(initContext));
    }
    if (useGlobalErrorHandler) {
        registerErrorHandlers(initContext, forceGlobalErrorHandler);
    }
    initContext.redact = createRedact(initContext, redactors);
    //
    // All Omit<LogContext> props should be filled by this point
    //
    const context = initContext;
    context.logger = {
        handler: wrapHandler(context),
        setMinimumLogLevel: wrapSetMinimumLogLevel(context),
        events: context.events,
        setKey: wrapSetKey(context),
        ...createLogger(context),
    };
    return context.logger;
}
exports.Logger = Logger;
function createLogger(logContext) {
    const severityLog = wrapLog(logContext);
    return {
        createSubLogger: wrapSubLogger(logContext),
        info: severityLog('info'),
        warn: severityLog('warn'),
        error: severityLog('error'),
        debug: severityLog('debug'),
    };
}
function wrapSubLogger(logContext) {
    return (name) => createLogger({
        ...logContext,
        // Reference parent's minimumLogLevel so that it cascades down
        get minimumLogLevel() {
            return logContext.minimumLogLevel;
        },
        contextPath: [...logContext.contextPath, name],
        // TODO: Figure out why you were doing this
        // logger: undefined
    });
}
function wrapHandler(logContext) {
    return (handler) => async (lambdaEvent, lambdaContext) => {
        // Create initial values from context
        setMdcKeys(logContext, lambdaEvent, lambdaContext);
        // Attach logger to context
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        lambdaContext.logger = logContext.logger;
        if (logContext.events)
            logContext.events.emit('beforeHandler', lambdaEvent, lambdaContext);
        const result = handler(lambdaEvent, lambdaContext);
        if (!isPromise(result)) {
            // log the error first
            throw new Error('Logger wrapped a handler function that did not return a promise. Lambda logger only supports async handlers.');
        }
        return result;
    };
}
function wrapSetKey(logContext) {
    return (key, value) => {
        if (reservedKeys.includes(key)) {
            throw new Error(`"${key}" is a reserved logger key.`);
        }
        logContext.keys.set(key, value);
    };
}
function setMdcKeys(logContext, lambdaEvent, lambdaContext) {
    let traceIndex = 0;
    logContext.logger.setKey('traceId', lambdaContext.awsRequestId);
    logContext.logger.setKey('date', getFormattedDate);
    logContext.logger.setKey('appName', lambdaContext.functionName);
    logContext.logger.setKey('version', lambdaContext.functionVersion);
    logContext.logger.setKey('apigTraceId', (lambdaEvent &&
        lambdaEvent.requestContext &&
        lambdaEvent.requestContext.requestId) ||
        (lambdaContext &&
            lambdaContext.requestContext &&
            lambdaContext.requestContext.requestId));
    logContext.logger.setKey('traceIndex', () => traceIndex++);
}
function wrapLog(logContext) {
    return (severity) => (...args) => writeLog(logContext, severity, ...args);
}
function writeLog(logContext, severity, ...args) {
    if (!canLogSeverity(logContext, severity))
        return;
    const logMessage = getLogMessage(logContext, severity, ...args);
    switch (severity) {
        case 'warn':
            system.console.warn(logMessage);
            return;
        case 'error':
            system.console.error(logMessage);
            return;
        case 'debug':
            system.console.debug(logMessage);
            return;
        case 'info':
        default:
            system.console.log(logMessage);
    }
}
function getLogMessage(logContext, severity, ...args) {
    const keys = Object.fromEntries(Array.from(logContext.keys.entries()).map(([key, value]) => [
        key,
        typeof value === 'function' ? value() : value,
    ]));
    return logContext.redact(logContext.formatter({ keys, contextPath: logContext.contextPath }, severity, ...args));
}
function canLogSeverity(logContext, severity) {
    if (!types_1.severities.includes(severity))
        throw new Error('Unable to log, illegal severity: ' + severity);
    return !(logContext.minimumLogLevel &&
        severity &&
        types_1.logLevels.indexOf(severity.toUpperCase()) <
            types_1.logLevels.indexOf(logContext.minimumLogLevel));
}
function wrapSetMinimumLogLevel(logContext) {
    return (level) => {
        if (types_1.logLevels.indexOf(level) === -1) {
            throw new Error('Illegal log level value: ' + level);
        }
        logContext.minimumLogLevel = level;
    };
}
function formatMessageItem(message) {
    if (typeof message === 'string')
        return message;
    return (0, fast_safe_stringify_1.default)(message, undefined, 2);
}
function JsonFormatter(context, severity, ...args) {
    const log = {
        ...context.keys,
    };
    log.message =
        args.length === 1
            ? formatMessageItem(args[0])
            : args.map(formatMessageItem).join(' ');
    log.severity = severity.toUpperCase();
    const subLogPath = getLogPath(context.contextPath);
    log.contextPath = subLogPath || undefined;
    // put the un-annotated message first to make cloudwatch viewing easier
    // include the MDC annotaed message after with log delimiter to enable parsing
    return `${subLogPath ? ` ${subLogPath} ` : ' '}${log.message} |\n ${withDelimiter((0, fast_safe_stringify_1.default)(log))}`;
}
function TestFormatter(context, severity, ...args) {
    const subLogPath = getLogPath(context.contextPath);
    const message = args.length === 1
        ? formatMessageItem(args[0])
        : args.map(formatMessageItem).join(' ');
    return `${severity.toUpperCase()}${subLogPath ? ` ${subLogPath} ` : ' '}${message}`;
}
function withDelimiter(message) {
    return exports.LOG_DELIMITER + message + exports.LOG_DELIMITER;
}
function getLogPath(paths) {
    return paths.join('.');
}
// Redaction
//
function createRedact(logContext, redactors) {
    logContext.redactors = redactors.map((redactor) => {
        if (typeof redactor === 'string')
            return (0, strings_1.stringRedactor)(redactor);
        else if (redactor instanceof RegExp)
            return (0, strings_1.regexRedactor)(redactor);
        else if (typeof redactor === 'function')
            return redactor;
        throw new Error(`Redactor type not supported: ${redactor}`);
    });
    return (value) => logContext.redactors.reduce((val, redactor) => redactor(val), value);
}
function bearerRedactor(logContext) {
    const tokens = [];
    const bearerRegex = /Bearer ([A-Za-z0-9-_]+\.[A-Za-z0-9-_]+\.[A-Za-z0-9-_]*)/; // eslint-disable-line no-useless-escape
    // Only keep the token for the current handler invocation
    if (logContext.events) {
        logContext.events.on('beforeHandler', () => {
            clearArray(tokens);
        });
    }
    return (val) => {
        // Don't handle empty or non-string values
        if (!val || typeof val !== 'string')
            return val;
        // Redact early, or...
        const token = tokens.find((f) => val.includes(f));
        if (token)
            return (0, strings_1.redact)(val, token);
        // match the token
        const match = val.match(bearerRegex);
        if (!match)
            return val;
        const bareToken = match[1];
        // When not using the handler wrapper, "tokens" grows unbounded
        // Don't keep more than 5 tokens
        // Since each token is the full and bare entry, length = 10
        if (tokens.length > 10) {
            // Remove the last match-pair
            tokens.splice(0, 2);
        }
        // Store the match pair
        tokens.push(val, bareToken);
        return (0, strings_1.redact)(val, match[0]);
    };
}
// Error Handling
//
function registerErrorHandlers(logContext, force) {
    if (!force && isInTestMode(logContext.testMode)) {
        return;
    }
    clearLambdaExceptionHandlers();
    logContext.globalErrorHandler = (err, promise) => {
        if (promise) {
            logContext.logger.error('unhandled rejection (this should never happen!)', err.stack);
        }
        else {
            logContext.logger.error('uncaught exception (this should never happen!)', err.stack);
        }
        // eslint-disable-next-line no-process-exit
        system.process.exit(1);
    };
    system.process.on('uncaughtException', logContext.globalErrorHandler);
    system.process.on('unhandledRejection', logContext.globalErrorHandler);
}
let hasAlreadyClearedLambdaHandlers = false;
function clearLambdaExceptionHandlers() {
    if (hasAlreadyClearedLambdaHandlers) {
        throw new Error('tried to setup global handlers twice. You cannot construct two Loggers with "useGlobalErrorHandler"');
    }
    // Taken from: https://gist.github.com/twolfson/855a823cfbd62d4c7405a38105c23fd3
    // DEV: AWS Lambda's `uncaughtException` handler logs `err.stack` and exits forcefully
    //   uncaughtException listeners = [function (err) { console.error(err.stack); system.process.exit(1); }]
    //   We remove it so we can catch async errors and report them to Rollbar
    if (system.process.listeners('uncaughtException').length !== 1) {
        throw new Error('Logger Assertion Failed: uncaughtException does not have 1 listener(s)');
    }
    system.process.removeAllListeners('uncaughtException');
    if (system.process.listeners('unhandledRejection').length !==
        system.UNHANDLED_REJECTION_LISTENERS) {
        throw new Error(`Logger Assertion Failed: unhandledRejection does not have ${system.UNHANDLED_REJECTION_LISTENERS} listener(s)`);
    }
    system.process.removeAllListeners('unhandledRejection');
    hasAlreadyClearedLambdaHandlers = true;
}
// Utilities
function isInTestMode(testMode) {
    // Override with context
    if (testMode !== undefined)
        return testMode;
    // Check environment
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const isMocha = global.it !== undefined;
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const isJest = global.jest !== undefined;
    const isTape = hasModuleLoaded('tape');
    // console.log('has tape loaded', isTape)
    const env = system.process.env.TEST ||
        system.process.env.test ||
        system.process.env.ci ||
        system.process.env.CI;
    // console.log('load', env, isMocha, isJest, isTape)
    return env || isMocha || isJest || isTape;
}
function hasModuleLoaded(moduleName) {
    try {
        return !!require.cache[require.resolve(moduleName)];
    }
    catch (e) {
        if (/Cannot find module/.test(e.toString()))
            return false;
        throw e;
    }
}
function pad(n) {
    return n < 10 ? '0' + n : n;
}
function getFormattedDate() {
    return formatRfc3339(new Date(Date.now()));
}
function formatRfc3339(d) {
    return `${d.getUTCFullYear()}-${pad(d.getUTCMonth() + 1)}-${pad(d.getUTCDate())}T${pad(d.getUTCHours())}:${pad(d.getUTCMinutes())}:${pad(d.getUTCSeconds())}Z`;
}
function isPromise(promise) {
    return (!!promise &&
        promise.then !== undefined &&
        typeof promise.then === 'function');
}
function clearArray(arr) {
    arr.length = 0;
}
